{ UseExcp.pas  v3.0  Copyright  5/94  Adolfo Di Mare }

{$I ADT-ADH.inc}

PROGRAM UseExcp;

{ RESULTS
  This program demonstrates how to use the Except.pas
  exception package.

  The program has been decorated with arrows to show
  how an exception is catched by the exception
  handling mechanism. The examples include catching
  run time fatal conditions, like division by zero or
  program stack overflow, and also examples of
  programmer generated exceptions through the Throw()
  and Signal() verbs.

  The interactive debugger can be used to see how the
  program works, but break points should be manually
  inserted in the code. Otherwise the program will
  run wild because single stepping does not work when
  an exception occurs. To catch the exception while
  using the debugger it is necessary to put a
  breakpoint in the ELSE clause of each exception
  context.                                             }

{ PROGRAMMER                         Adolfo Di Mare    }

{ ACKKNOWLEDGEMENTS                  David Chavez
                                     Gustavo Sanabria
                          Break.pas: Scott Bussinger   }


{ COPYRIGHT
  This code is copyrighted by Adolfo Di Mare. You are
  free to use it and distribute it. You cannot sell
  it, or blame Adolfo for any problems that software
  components containing this code could produce.
  You are on your own.                                 }

USES
  Dos,
  Except,  { There is no need to place Except as }
  Crt      { the last unit in the main program   }

  {$IFDEF Ver70}
     {$DEFINE Version_Supported}
   ,CBreak  { Use it, if you have it...          }
     {$DEFINE Version_Supported}
  {$ENDIF}
  ;

{ UNIT CBreak.pas (formerly Break.pas)
  A unit to allow Control-Break to interrupt program
  execution.

  Version 1.00 -  1/02/1987 - First general release

  Scott Bussinger
  Professional Practice Systems
  110 South 131st Street
  Tacoma, WA  98444
  Compuserve 72247,2671                              }

VAR
  GLOBAL : RECORD
    ExitSave : Except.PExitProc;
  END;

{$F+}
PROCEDURE ExitProcedure;
{ RESULTS
  This procedure is chained to the exit procedures
  using the Register_Exit_Procedure() verb. It
  demonstrates how the programmer can use exit
  procedures in his program without disturbing the
  excepting handling mechanism.
  - An exit procedure is one that gets exectuted
    before the program terminates. Borland defined
    and ad-hoc mechanism for this which is described
    in the Turbo manual.                             }
BEGIN { ExitProcedure }
  {$IFOPT F-}
    Error: you must compile this procedure using
           the FAR memory model.
  {$ENDIF}
  { This is ExitProcedure's body }
  Write  ('ExitProcedure [');
  WriteLn('System.ExitCode = ', System.ExitCode,']');

  { Exit code: chain to other exit procedures }
  PExitProc(System.ExitProc) := GLOBAL.ExitSave;
  {Throw(18);}
END;  { ExitProcedure }
{$F-}

PROCEDURE Cicle({+} i:INTEGER);
{ RESULTS
  This procedure calls itself recursively,
  but it never returns. Instead, it uses
  Throw() to return a value.
  - The procedure returns immediatly,
    without scaling through all the
    recursive calls. That's what Throw()
    is good for!.                          }
BEGIN { Cicle }
  Write('==> call Cicle(', i:1, ')');
  IF i = 6 THEN BEGIN
    Write('==> Cicle(',i:1,')::Throw');
    Throw(i);
    WriteLn('Never executed!');
  END
  ELSE IF i = 8 THEN BEGIN
    Write('==> Cicle(',i:1,')::Signal');
    Signal(i);
    WriteLn('Never executed!');
  END
  ELSE IF i>0 THEN BEGIN
    Cicle(i-1);
    WriteLn('Never executed!');
  END
  ELSE BEGIN
    Write('==> Cicle(',i:1,')::Throw');
    Throw(i);
    WriteLn('Never executed!');
  END;

  { The following code will never be executed, }
  { because there is a Throw() or Signal()     }
  { for every posible IF condition.            }

  WriteLn('Never executed!');
  WriteLn('Returning from Cicle(', i:1, ')' );
END;  { Cicle }


PROCEDURE Deep_Throat({+} n : INTEGER);
{ RESULTS
  Shows what happens when the program stack
  gets exhausted by a recursive procedure.  }
{ REMARKS
  Try watching the action using the IDE
  debugger: Alt-W-k [Ctrl-F3] will show how
  deep the stack grows.                     }
VAR
  a: ARRAY[1..20] OF REAL;
BEGIN { Deep_Throat }
  IF n = 2639 THEN BEGIN
    Write('!');
    a[1] := 0.0;
  END;

  IF n MOD 100 = 0 THEN BEGIN
    Write('.');
  END;

  Deep_Throat(n+1);
END;  { Deep_Throat }

PROCEDURE Throw_vs_Signal;
{ RESULTS
  Shows why Throw() is different from Signal(). }
VAR
  i : INTEGER;
BEGIN { Throw_vs_Signal }
  i := 0;
  WriteLn; Write(' ', i:1);
  IF NOT Exception THEN BEGIN
    INC(i); Write(' ', i:1);
    IF NOT Exception THEN BEGIN
      INC(i); Write(' ', i:1);
      IF NOT Exception THEN BEGIN
        INC(i); Write(' ', i:1, ' ');

        IF i = 3 THEN BEGIN
          { Invoke the closest exception handler }
          Throw(25);                { => => => => => =>}
        END;                                        {||}
                                                    {||}
        Leave_Exception_Context;                    {\/}
      END                                           {  }
      ELSE BEGIN                                    {||}
        WriteLn('Level #3');                        {\/}
        CASE Catch OF               { <= <= <= <= <= <=}

          25: Write('Throw(25) received');  { !!!!!! }

        ELSE
          WriteLn(
            '!Throw(25)==>Throw(',
             Exception_Code, ')' );
        END;

        { => => => =>  BYE BYE BYE Throw_Signal() !!! }

        Signal(9);   { returns to the calling program }

        { Signal() will never branch to an exception  }
        { handler within the same procedure: it will  }
        { always branch to an exception context in an }
        { invoking procedure.                         }

        { Hopefully, an "invoking procedure" is part  }
        { of a "higher level" software layer.         }

      END;

      Leave_Exception_Context;
    END
    ELSE BEGIN
      WriteLn('Level #2');
      WriteLn('Catch=', Catch);
      Re_Throw;
    END;

    Leave_Exception_Context;
  END
  ELSE BEGIN
    WriteLn('Level #1');
    WriteLn('Catch=', Catch);
    Re_Throw;
  END;
END;  { Throw_vs_Signal }


PROCEDURE Forget_Leave_Exception({+} n : INTEGER);
{ RESULTS
  Creates a lot of exception contexts and forgets
  to release them. Fortunately, the calling
  procedure can tell the error, but this can not
  always be the case.                            }
VAR
  i : INTEGER;
BEGIN { Forget_Leave_Exception }

  WriteLn('I will create some exception contexts ');
  Write  ('==> I will NOT release them!!!  ');

  FOR i := 1 TO n DO BEGIN
    IF NOT Exception THEN BEGIN
      Write('+');

      { Forget Leave_Exception_Context; on purpose }
    END
    ELSE BEGIN
      { There are no exceptions for these handlers }
      Write('-');
    END;
  END;

  WriteLn('  (',n:1,')');

  { The value for Except.stack.top is n. This }
  { means that at this point, the exception   }
  { handler believes that there are n levels  }
  { of nested exceptions.                     }

END;  { Forget_Leave_Exception }

PROCEDURE Exception_without_Leave_Exception;
{ RESULTS
  Shows how some defensive programming used
  within procedure Exception can help the
  careless programmer when he "forgets" to invoke
  the Leave_Exception_Context verb.               }

{ WHY
  Why is it posible for Exception to fix its
  stack?                                          }

{ BECAUSE
  The answer is simple: as the programmer forgot
  to invoke the Leave_Exception_Context verb
  inside a procedure, Exception can tell that the
  depth of the program stack does not coincide
  with the values stored in the exception stack.

  Exception isn't fool proof: there are ocassions
  when a missing Leave_Exception_Context will not
  be detected. In these cases, the program will
  crash sooner or later, and it will do so in a
  most-unmissgiven-way!!!                         }

VAR
  initial,           { # of current excp. contexts  }
  broken,            { # of "broken" excp. contexts }
  fixed   : INTEGER; { should match initial         }

BEGIN { Exception_without_Leave_Exception }

  initial := Exception_Depth;   { ==> initial := 0; }

  IF NOT Exception THEN BEGIN
    Forget_Leave_Exception(
      Except.Exception_Stack_Size + 8
    );

    WriteLn('As Except.pas is dumb, ',
            'this gets executed');

    Leave_Exception_Context;
  END
  ELSE BEGIN
    { Exception handler }
    WriteLn('I will do nothing now');
  END;

  broken := Exception_Depth;

  { Upon return from procedure Forget_Leave_Exception }
  { function Exception is invoked. It detects that    }
  { the exception stack is corrupted, an hence throws }
  { the UnReleased_Exception_Context exception.       }

  IF NOT Exception THEN BEGIN
    WriteLn('This never gets executed');

    Leave_Exception_Context;
  END
  ELSE BEGIN
    { Exception handler }
    CASE Catch OF
      UnReleased_Exception_Context : BEGIN
        Write  ('====> I cleaned up the mess!!!');
        WriteLn(' [but I should fix the program]');
      END;
      Exception_Stack_Overflow : BEGIN
        Write  ('====> Too many nested exceptions');
        WriteLn(' [but this never gets executed]');
      END;
    ELSE
      WriteLn('Exception (', Exception_Code, ')');
      Re_Throw;
    END { CASE }
  END;

  fixed := Exception_Depth;

  IF initial <> fixed THEN BEGIN
    WriteLn('Couldn''t fix the exception stack');
  END;

END;  { Exception_without_Leave_Exception }


PROCEDURE WORK;
VAR
  clean : ARRAY[0..5] OF BYTE;
  a : REAL;
  i : INTEGER;
BEGIN { WORK }

  ClrScr;
  WriteLn('************');

  IF NOT Exception THEN BEGIN

    FOR i := 0 TO SizeOf(clean) DIV SizeOf(clean[0])
    DO BEGIN
      clean[i] := 0;      { Range_Check_Error ==> => =>}
    END;                                            {||}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}
      Range_Check_Error : BEGIN
        WriteLn('clean[',i:1,'] is out of bounce');
        i := 0;
      END;
    ELSE
      EXIT;
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN

    a := 123456789.0;
    IF NOT Exception THEN BEGIN

      a := Exp(100000); { Floating_Point_Overflow => =>}
      Cicle(10);        { never gets executed...  } {||}
                                                    {||}
      Leave_Exception_Context;                      {\/}
    END                                             {  }
    ELSE BEGIN                                      {||}
      { Exception handler }                         {\/}
      CASE Catch OF                 { <= <= <= <= <= <=}
        Program_Stack_Overflow :
          a := 0;
        8  :
          WriteLn('Ignored by Signal()');
        211:
          a := 1.e+37;

      ELSE
        Re_Throw;     { => => => => => => => => => =>||}
      END { CASE }                                  {||}
    END; { Exception }                              {\/}
                                                    {  }
    { This block of code is never executed }        {||}
    IF NOT Exception THEN BEGIN                     {\/}
      { very-many-muchos-muchos-calculos }          {  }
      a := Ln(-1) * Exp(1000000);                   {||}
      a := Exp(100000);                             {\/}
      a := Exp(100000);                             {  }
      Leave_Exception_Context;                      {||}
    END                                             {\/}
    ELSE BEGIN                                      {  }
      { Don't care Exception handler... }           {||}
      a := 0;                                       {\/}
    END; { Exception }                              {  }
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}
      1: WriteLn('ONE');
      2: WriteLn('TWO');
      6: WriteLn('SIX: First Exception block');
      8: WriteLn('EIGHT: Throw()');
      Floating_Point_Overflow,
      Floating_Point_Underflow,
      Invalid_Floating_Point_Opr : BEGIN
        Write  ('Overflow found ');           { !!!!!! }
        a := 0;
        WriteLn(' a := ', a:3:1);
      END;
    ELSE
      Re_Throw;
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN

    a := 123456789.0;
    IF NOT Exception THEN BEGIN
      a := 2 * a;

      Leave_Exception_Context;
    END
    ELSE BEGIN
      { Exception handler }
      CASE Catch OF
        Program_Stack_Overflow :
          a := 0;
        8  :
          WriteLn('Ignored by Signal()');
        211:
          a := 1.e+37;
      ELSE
        Re_Throw;
      END { CASE }
    END; { Exception }

    Leave_Exception_Context;
  END
  ELSE BEGIN
    { Exception handler }
    CASE Catch OF
      1: WriteLn('ONE');
      2: WriteLn('TWO');
      6: WriteLn('SIX: First Exception block');
      8: WriteLn('EIGHT: Throw()');
      Floating_Point_Overflow: BEGIN
        WriteLn('Overflow found');
        a := 0;
      END;
    ELSE
      Re_Throw;
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN
    Cicle(7);                    { => => => => => => =>}
    Cicle(5);   { never executed }                  {||}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF             { <= <= <= <= <= <= <= <=}

      6: WriteLn('(',Exception_Code:1,')');   { !!!!!! }

      7: WriteLn('SEVEN');
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN
    Cicle(9);                    { => => => => => => =>}
    Cicle(5);   { never executed }                  {||}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                   { <= <= <= <= <= <=}
      8: WriteLn('(',Exception_Code:1,')');   { !!!!!! }
    ELSE
      Re_Throw;
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN
    Cicle(2);                       { => => => => => =>}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                   { <= <= <= <= <= <=}

      0: WriteLn('(', Exception_Code:1, ')'); { !!!!!! }
         { Yes: zero can be an exception code... }

      7: WriteLn('SEVEN');
    END { CASE }
  END; { Exception }

  IF NOT Exception THEN BEGIN
    Throw_vs_Signal;             { => => => => => => =>}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}

      8: WriteLn('(', Exception_Code:1, ')');

      9: WriteLn(' ==> Signal(NINE)');        { !!!!!! }

    END { CASE }
  END; { Exception }


  { After quite a while, we decide to chain an exit }
  { procedure into the exit procedure chain.        }
  { - Don't do it by hand; you would break Except!  }

  Register_Exit_Procedure(
    GLOBAL.ExitSave,       { Exit Chain     }
    ExitProcedure          { Exit procedure }
  );


  WriteLn('a := ', a:1:1,'         i := ', i:1);

  Exception_without_Leave_Exception;

  IF NOT Exception THEN BEGIN    { => => => => => => =>}
    Leave_Exception_Context;                        {  }
  END                                               {||}
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}
      UnReleased_Exception_Context:
        WriteLn ('Exception stack cleared!'); { !!!!!! }
    ELSE
      i := 1;
    END { CASE }
  END; { Exception }

  { The following code won't work within the IDE; }
  { try it from the command line.                 }
  IF NOT Exception THEN BEGIN
      { This FOR takes a while to execute...  }
      { Try Ctrl-Break to see how Break works }
      FOR i := 1 TO 20 DO BEGIN
         Delay(100);   { Ctrl-Break detected   => => =>}
      END;                                          {||}
      i := MaxInt;     { No Ctrl-Break }            {||}
                                                    {\/}
    Leave_Exception_Context;                        {  }
  END                                               {||}
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}
      Ctrl_Break: BEGIN
        WriteLn ('Ctrl-Break Detected!');    { !!!!!!! }
        i := 0;
      END;
    ELSE
      i := 1;
    END { CASE }
  END; { Exception }

  CASE i OF
    MaxInt: BEGIN
      Write  ('Ctrl-Break cannot be detected ');
      WriteLn('within the IDE');
    END;
    0 : BEGIN { Do Nothing } END;
    ELSE
      WriteLn('Unhandled exception ', Exception_Code);
  END;


  IF NOT Exception THEN BEGIN
    Write('(', Exception_Code:1, ')');
    Deep_Throat(0);              { => => => => => => =>}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {\/}
    CASE Catch OF                { <= <= <= <= <= <= <=}
      Program_Stack_Overflow:
        WriteLn (' Program_Stack_Overflow'); { !!!!!!! }
      Ctrl_Break:
        WriteLn ('Again Ctrl-Break!');       { ??????? }
    ELSE
        WriteLn('Bombed! (', Exception_Code:1, ')');
    END { CASE }

  END; { Exception }


  IF NOT Exception THEN BEGIN

    { This section of code checks whether variable   }
    { clean[] was overwritten in the program. Should }
    { this had happened, it could mean that when the }
    { program stack overflowed, the bottom of the    }
    { stack got overwritten.                         }
    FOR i := 0 TO SizeOf(clean) DIV SizeOf(clean[0])
    DO BEGIN                              { => => => =>}
      IF clean[i] <> 0 THEN BEGIN                   {||}
        Throw(Program_Stack_Overflow);              {\/}
      END;                                          {  }
    END;                                            {||}
                                                    {||}
    Leave_Exception_Context;                        {\/}
  END                                               {  }
  ELSE BEGIN                                        {||}
    { Exception handler }                           {||}
    CASE Catch OF                                   {\/}
      Range_Check_Error : BEGIN  { <= <= <= <= <= <= <=}
        WriteLn('clean[',i:1,'] is out of bounce');
        i := 0;
      END;
      Program_Stack_Overflow: BEGIN
        WriteLn('A Program_Stack_Overflow destroyed me!');
      END;
    END { CASE }
  END; { Exception }


END;  { WORK }


BEGIN { UseExcp }
  WORK;
END.  { UseExcp }

{ EOF: UseExcp.pas }
